import asyncio
from typing import List
from urllib.parse import urljoin, urlparse

from bs4 import BeautifulSoup
from playwright.async_api import TimeoutError as PlaywrightTimeoutError

from pyba.utils.common import url_entropy
from pyba.utils.load_yaml import load_config
from pyba.utils.structure import CleanedDOM

general_config = load_config("general")
config = load_config("extraction")[
    "general"
]  # This means we're referring to the general extraction class


class GeneralDOMExtraction:
    """
    Given the DOM from the URL, this class provides functions to extract it properly

    1. Extract all the hyperlinks from it
    2. Extract all input_fields from it (basically all fillable boxes)
    3. Extract all the clickables from it
    4. Extract all the actual text from it (we don't have to do any OCR for this!)

    Note that extracing all clickable elements might get messy so we'll use that only when
    the total length is lower than a certain threshold.

    This is the general extraction. Some specific websites require a different way to do this. Checkout the README for more.
    """

    def __init__(
        self,
        html: str,
        body_text: str,
        elements: str,
        base_url: str = None,
        clickable_fields_flag: bool = False,
    ) -> None:
        """
        We'll take the entire dom, the text_body and the elements for sure
        """

        self.html = html
        self.body_text = body_text
        self.elements = elements
        self.base_url = base_url

        self.clickable_fields_flag = clickable_fields_flag
        # For testing fields
        self.test_value = general_config["main_engine_configs"]["input_field_test_value"]

    def _extract_clickables(self) -> List[dict]:
        soup = BeautifulSoup(self.html, "html.parser")
        candidates = []

        for tag in soup.find_all(
            list(config["extraction_configs"]["clickables"]["clickable_field_selectors"])
        ):
            if tag.name == "a":
                href = tag.get("href", "").strip().lower()
                if (
                    not href
                    or href
                    in set(
                        config["extraction_configs"]["clickables"][
                            "invalid_selector_field_hyperlinks"
                        ]
                    )
                    or href.startswith("javascript:")
                    or href.startswith("#")
                ):
                    continue
            candidates.append(tag)

        for tag in soup.find_all("input"):
            t = tag.get("type", "").lower()
            if t in tuple(
                config["extraction_configs"]["clickables"]["valid_button_types_for_clickables"]
            ):
                candidates.append(tag)

        candidates += soup.find_all(attrs={"onclick": True})
        candidates += soup.find_all(
            attrs={"role": lambda v: v and v.lower() in ("button", "link")}
        )
        candidates += soup.find_all(attrs={"tabindex": True})

        seen = set()
        results = []
        for el in candidates:
            key = (el.name, str(el))
            if key in seen:
                continue
            seen.add(key)

            href = el.get("href")
            onclick = el.get("onclick")
            role = el.get("role")
            tabindex = el.get("tabindex")
            text = el.get_text(strip=True)

            if not (text or href or onclick):
                continue

            if href and self.base_url:
                href = urljoin(self.base_url, href)

            # Not sure how junk these are but dropping them for now to avoid explosion of context
            junk_keywords = config["extraction_configs"]["clickables"]["junk_keywords"]
            if any(k in text.lower() for k in junk_keywords):
                continue

            results.append(
                {
                    "tag": el.name,
                    "text": text,
                    "href": href,
                    "onclick": onclick,
                    "role": role,
                    "tabindex": tabindex,
                    "outer_html": str(el)[:1000],
                }
            )

        # Cleaning this up based on outer_html

        cleaned = []
        for el in results:
            data = {k: v for k, v in el.items() if v and k != "outer_html"}
            cleaned.append(data)

        return cleaned

    def _extract_href(self) -> List[str]:
        soup = BeautifulSoup(str(self.html), "html.parser")
        hrefs = [a["href"].strip() for a in soup.find_all("a", href=True)]

        clean_hrefs = []
        for href in hrefs:
            href_lower = href.lower()

            # If we do a raw extraction, all this junk will make it through
            if (
                not href_lower
                or href_lower
                in set(
                    config["extraction_configs"]["clickables"]["invalid_selector_field_hyperlinks"]
                )
                or href_lower.startswith("javascript:")
                or href_lower.startswith("#")
            ):
                continue

            # Have to figure out a bettery way to handle these
            # In the case where we do have relative URLs, we just make it absolute
            full_url = urljoin(self.base_url, href)

            if any(
                x in href_lower
                for x in list(config["extraction_configs"]["hyperlinks"]["links_to_avoid"])
            ):
                continue

            # Skipping any other type of URL. Will have to make this a confugrable parameter
            parsed = urlparse(full_url)
            if parsed.scheme not in set(
                config["extraction_configs"]["hyperlinks"]["valid_schemas"]
            ):
                continue

            clean_hrefs.append(full_url)

        # Before moving forward, we can filter them based on entropy
        """
        From what I noticed, big sites like amazon or facebook, have lots and lots of URLs and most of them
        contain a bunch of random strings for IDs and whatnot. We don't want that because that gives no
        additional context to the model regarding what the href is actually for!

        So, we'll strip those based on entropy. I noticed from multiple scans that the shannon entropy for
        such URLs are almost always greater than 5. Its interesting...
        """

        output = [href for href in clean_hrefs if url_entropy(href) < 5.0]
        return output

    async def _extract_all_text(self) -> List:
        lines = self.body_text.split("\n")
        non_empty_lines = [line.strip() for line in lines if line.strip()]
        return non_empty_lines

    async def _extract_input_fields(
        self,
        known_fields: List = None,
    ) -> List[dict]:
        """
        Extracts and verifies fillable input fields. We're passing to it all the valid fields we already know
        of so that it caches it and doesn't mess with the actual DOM during execution.

        Args:
            known_fields : List[Dict], optional
            Previously detected valid fields (to avoid duplicate fills).

        Returns:
            List[Dict]
                List of valid fillable fields with tag/type/id/name/selector info.
        """

        valid_fields = [] if known_fields is None else known_fields.copy()
        seen_selectors = {f["selector"] for f in valid_fields if f.get("selector")}

        for el in self.elements:
            try:
                is_visible = await el.is_visible()
                is_enabled = await el.is_enabled()
                if not (is_visible and is_enabled):
                    continue

                tag = (await el.evaluate("e => e.tagName.toLowerCase()")).strip()
                input_type = await el.get_attribute("type")
                input_type = (input_type or "text").lower().strip()
                existing_value = await el.input_value() if tag in ["input", "textarea"] else ""

                if tag not in set(config["extraction_configs"]["input_fields"]["valid_tags"]):
                    continue
                if input_type in set(
                    config["extraction_configs"]["input_fields"]["invalid_input_types"]
                ):
                    continue

                field_info = {
                    "tag": tag,
                    "type": input_type,
                    "id": await el.get_attribute("id"),
                    "name": await el.get_attribute("name"),
                    "placeholder": await el.get_attribute("placeholder"),
                    "aria_label": await el.get_attribute("aria-label"),
                    "selector": None,
                }

                if field_info["id"]:
                    selector = f"#{field_info['id']}"
                elif field_info["name"]:
                    selector = f"{field_info['tag']}[name='{field_info['name']}']"
                elif field_info["placeholder"]:
                    selector = f"{field_info['tag']}[placeholder='{field_info['placeholder']}']"
                elif field_info["aria_label"]:
                    selector = f"{field_info['tag']}[aria-label='{field_info['aria_label']}']"
                else:
                    selector = f"{field_info['tag']}:nth-of-type(unknown)"

                field_info["selector"] = selector

                # If the field is already filled or we've seen it before,
                # Then we can skip it
                if selector in seen_selectors:
                    continue

                if existing_value and existing_value.strip():
                    valid_fields.append(field_info)
                    seen_selectors.add(selector)
                    continue

                # Filling it and seein
                await el.fill(self.test_value, timeout=1500)
                await asyncio.sleep(0.05)
                new_value = await el.input_value()

                if self.test_value in new_value:
                    valid_fields.append(field_info)
                    seen_selectors.add(selector)
                    await el.fill("")  # clearing the field after testing it
                else:
                    pass  # ignore false positives

            except PlaywrightTimeoutError:
                continue
            except Exception:
                continue

        return valid_fields

    async def extract(self) -> CleanedDOM:
        """
        Runs all extraction functions and returns a unified cleaned_dom dictionary.

        Returns:
            CleanedDOM: A dataclass containing hyperlinks, input fields, clickable fields, and text content.
        """
        cleaned_dom = CleanedDOM()

        try:
            cleaned_dom.hyperlinks = self._extract_href()
        except Exception as e:
            cleaned_dom.hyperlinks = []
            print(f"Failed to extract hyperlinks: {e}")

        try:
            if self.clickable_fields_flag:
                cleaned_dom.clickable_fields = self._extract_clickables()
            else:
                cleaned_dom.clickable_fields = self._extract_clickables()[:10]
                # This is taking way too many tokens so restricting the total number.
                # There has to be a better way to do this!
        except Exception as e:
            cleaned_dom.clickable_fields = []
            print(f"Failed to extract clickables: {e}")

        try:
            cleaned_dom.actual_text = await self._extract_all_text()
        except Exception as e:
            cleaned_dom.actual_text = []
            print(f"Failed to extract text: {e}")

        try:
            cleaned_dom.input_fields = await self._extract_input_fields()
        except Exception as e:
            cleaned_dom.input_fields = []
            print(f"Failed to extract input fields: {e}")

        return cleaned_dom



from dataclasses import dataclass, field
from typing import Optional, List

from pydantic import BaseModel


class PlaywrightAction(BaseModel):
    """
    The BaseModel for playwright automations

    Goal:
        This contains an exhaustive list of commands that playwright can execute. It
        will be filled accordingly by the LLM depending on the DOM recieved from playwright
        and the goal of the task.
    """

    goto: Optional[str]
    go_back: Optional[bool]
    go_forward: Optional[bool]
    reload: Optional[bool]

    click: Optional[str]
    dblclick: Optional[str]
    hover: Optional[str]
    fill_selector: Optional[str]
    fill_value: Optional[str]
    type_selector: Optional[str]
    type_text: Optional[str]
    press_selector: Optional[str]
    press_key: Optional[str]
    check: Optional[str]
    uncheck: Optional[str]
    select_selector: Optional[str]
    select_value: Optional[str]
    upload_selector: Optional[str]
    upload_path: Optional[str]

    scroll_x: Optional[int]
    scroll_y: Optional[int]
    wait_selector: Optional[str]
    wait_timeout: Optional[int]
    wait_ms: Optional[int]

    keyboard_press: Optional[str]
    keyboard_type: Optional[str]
    mouse_move_x: Optional[int]
    mouse_move_y: Optional[int]
    mouse_click_x: Optional[int]
    mouse_click_y: Optional[int]

    new_page: Optional[str]
    close_page: Optional[bool]
    switch_page_index: Optional[int]

    evaluate_js: Optional[str]
    screenshot_path: Optional[str]
    download_selector: Optional[str]


class PlaywrightResponse(BaseModel):
    actions: List[PlaywrightAction]


class OutputResponseFormat(BaseModel):
    """
    Output type for the model for direct response
    """

    output: str


@dataclass
class CleanedDOM:
    """
    Represents the cleaned DOM snapshot of the current browser page.

    Additional parameter for the youtube DOM extraction
    """

    hyperlinks: Optional[List[str]] = field(default_factory=list)
    input_fields: Optional[List[str]] = field(default_factory=list)
    clickable_fields: Optional[List[str]] = field(default_factory=list)
    actual_text: Optional[str] = None
    current_url: Optional[str] = None
    youtube: Optional[str] = None  # For YouTube based DOM extraction

    def to_dict(self) -> dict:
        return {
            "hyperlinks": self.hyperlinks,
            "input_fields": self.input_fields,
            "clickable_fields": self.clickable_fields,
            "actual_text": self.actual_text,
            "current_url": self.current_url,
            "youtube": self.youtube,
        }



from sqlalchemy import create_engine

from pyba.database.models import Base


class SQLiteHandler:
    """
    Class for handling SQLite table creation
    """

    def __init__(self, database_engine_configs):
        """
        Args:
                `database_engine_configs`: User supplied configurations imported from the Database class
        """

        self.database_engine_configs = database_engine_configs
        self.database_connection_string = self.database_engine_configs.database_connection_string

        self.engine = create_engine(
            self.database_connection_string, connect_args={"check_same_thread": False}
        )

    def sqlite_create_tables(self):
        """
        Method to create the SQLite tables
        """
        Base.metadata.create_all(self.engine)

    def setup(self):
        """
        Entrypoint for database and table creation
        """
        self.sqlite_create_tables()



import json
import re
from typing import List

from pyba.database import DatabaseFunctions


class CodeGeneration:
    """
    Create the full automation code used by the model

    - Requires the database to be populated with all the actions
    - Pulls action from the database and writes the script at a user location

    Args:
            `session_id`: The unique idenfier for this session
            `output_path`: Path to save the code to
            `database_funcs`: The Database instantiated by the user

    Probably rewritten a lot of the code I already have from the dispatcher. There must be a simpler way
    TODO: Clean up
    """

    def __init__(self, session_id: str, output_path: str, database_funcs: DatabaseFunctions):
        self.session_id = session_id
        self.output_path = output_path
        self.db_funcs = database_funcs

        # Mapping of action fields to their corresponding Playwright code method and format
        # The format string uses {selector} and {value} placeholders
        # For actions involving both selector and value, both are expected.
        # For navigation/single-value actions, only the value is used (as {value}).
        self.action_map = {
            "goto": (
                'page.goto("{value}")',
                1,
            ),  # (code_template, num_args: 1 for value, 2 for selector/value)
            "go_back": ("page.go_back()", 0),
            "go_forward": ("page.go_forward()", 0),
            "reload": ("page.reload()", 0),
            "click": ('page.click("{value}")', 1),
            "dblclick": ('page.dblclick("{value}")', 1),
            "hover": ('page.hover("{value}")', 1),
            "fill_selector": ('page.fill("{selector}", "{value}")', 2),  # Uses fill_value
            "type_selector": ('page.type("{selector}", "{value}")', 2),  # Uses type_text
            "press_selector": ('page.press("{selector}", "{value}")', 2),  # Uses press_key
            "check": ('page.check("{value}")', 1),
            "uncheck": ('page.uncheck("{value}")', 1),
            "select_selector": (
                'page.select_option("{selector}", "{value}")',
                2,
            ),  # Uses select_value
            "upload_selector": (
                'page.set_input_files("{selector}", "{value}")',
                2,
            ),  # Uses upload_path
            "scroll_x": (
                "page.mouse.wheel({value_x}, {value_y})",
                2,
                "scroll_y",
            ),  # Special handling in _parse_action_to_code
            "wait_selector": (
                'page.wait_for_selector("{value}", timeout={value_timeout})',
                2,
                "wait_timeout",
            ),  # Special handling
            "wait_ms": ("page.wait_for_timeout({value})", 1),  # Uses wait_ms
            "keyboard_press": ('page.keyboard.press("{value}")', 1),
            "keyboard_type": ('page.keyboard.type("{value}")', 1),
            "mouse_move_x": (
                "page.mouse.move({value_x}, {value_y})",
                2,
                "mouse_move_y",
            ),  # Special handling
            "mouse_click_x": (
                "page.mouse.click({value_x}, {value_y})",
                2,
                "mouse_click_y",
            ),  # Special handling
            # New pages/context actions are generally harder to script simply; using basic page-level equivalent
            "screenshot_path": ('page.screenshot(path="{value}")', 1),
            "evaluate_js": ('page.evaluate("{value}")', 1),
            "download_selector": (
                'with page.expect_download() as download_info:\n    page.click("{value}")\ndownload = download_info.value\ndownload.save_as(download.suggested_filename)',
                1,
            ),
            "new_page": (
                '# NOTE: Creating new pages is complex in a simple script\npage.context.new_page().goto("{value}")',
                1,
            ),  # Simplified, not strictly correct for current page state
            "close_page": ("page.close()", 0),
            "switch_page_index": (
                "# NOTE: Page switching is context-dependent\npage = page.context.pages[{value}]",
                1,
            ),  # Simplified assumption
        }

    def _get_run_actions(self) -> List:
        """
        Function to query the database and get the stack of all actions
        that have been performed up to the time the function was called
        """

        logs = self.db_funcs.get_episodic_memory_by_session_id(session_id=self.session_id)

        if logs and logs.actions:
            actions_list = json.loads(logs.actions)
            return actions_list
        return []

    def _parse_action_to_code(self, action_str: str) -> str:
        """
        Converts a single action string (e.g., 'goto="url" fill_selector=None...')
        into a Playwright code string.
        """
        # The first step is to parse the action string into a dictionary of key-value pairs
        # This pattern is robust to handle both 'key=value' and 'key="value"'
        action_data = {}

        # Replace quoted strings with a placeholder to avoid splitting on commas inside the value
        # A simpler approach given the specific format is to split by spaces and then by '='
        # The initial format of the action string is: key1=value1 key2=value2 ...
        parts = re.findall(r"(\w+)=([^ ]*)", action_str)

        for key, value in parts:
            # Strip quotes/None/None-like strings
            cleaned_value = value.strip().strip("'\"")
            if cleaned_value.lower() not in ("none", "false", ""):
                action_data[key] = cleaned_value

        code_lines = []

        # Check the action_map keys to find the one that is present (not None) in action_data
        for action_field, (template, num_args, *extra_field) in self.action_map.items():
            if action_field in action_data:
                # Navigation actions (e.g., go_back, reload)
                if num_args == 0:
                    code_lines.append(template)
                    break  # Assuming only one action per step right now

                elif num_args == 1:
                    if action_field == "download_selector":
                        code = template.replace('"{value}"', f'"{action_data[action_field]}"')
                        code_lines.extend(code.split("\n"))
                    else:
                        # Simple template replacement
                        code = template.format(value=action_data[action_field])
                        code_lines.append(code)
                    break

                # Two-argument actions (e.g., fill_selector/fill_value, scroll_x/scroll_y)
                elif num_args == 2:
                    # 1. Selector/Value pairs (e.g., fill_selector/fill_value, type_selector/type_text)
                    if action_field.endswith("_selector") or action_field == "wait_selector":
                        # Determine the corresponding value field
                        if action_field == "fill_selector":
                            value_field = "fill_value"
                        elif action_field == "type_selector":
                            value_field = "type_text"
                        elif action_field == "press_selector":
                            value_field = "press_key"
                        elif action_field == "select_selector":
                            value_field = "select_value"
                        elif action_field == "upload_selector":
                            value_field = "upload_path"
                        elif action_field == "wait_selector":
                            value_field = "wait_timeout"
                        else:
                            continue  # Should not happen

                        selector = action_data.get(action_field)
                        value = action_data.get(value_field)

                        if selector and value is not None:
                            # Handlin special case for wait_selector and wait_timeout
                            if action_field == "wait_selector" and "wait_timeout" in action_data:
                                timeout = action_data["wait_timeout"]
                                code = template.format(value=selector, value_timeout=timeout)
                            # Handling special case for fill_value being an empty string (is still a valid action)
                            elif (
                                action_field == "fill_selector"
                                and action_data.get("fill_value") == ""
                            ):
                                code = template.format(selector=selector, value="")
                            else:
                                code = template.format(selector=selector, value=value)

                            code_lines.append(code)
                            break

                    # 2. X/Y pairs (e.g., scroll_x/scroll_y)
                    elif (
                        action_field.endswith("_x")
                        and extra_field
                        and extra_field[0].endswith("_y")
                    ):
                        y_field = extra_field[0]
                        x_val = action_data.get(action_field)
                        y_val = action_data.get(y_field)

                        if x_val is not None or y_val is not None:
                            # Default to 0 if the other is None, as per Playwright
                            x_val = x_val if x_val is not None else 0
                            y_val = y_val if y_val is not None else 0

                            code = template.format(value_x=x_val, value_y=y_val)
                            code_lines.append(code)
                            break

        return (
            "\n".join(code_lines)
            if code_lines
            else f"# Action not supported or complete: {action_str}"
        )

    def generate_script(self):
        """
        Generates the full Playwright script from the sequence of actions and
        writes it to the output path.
        """

        actions_list = self._get_run_actions()

        # Boilerplate
        script_header = (
            "import time\n"
            "from playwright.sync_api import sync_playwright\n\n"
            "def run_automation():\n"
            "    with sync_playwright() as p:\n"
            "        browser = p.chromium.launch(headless=False)\n"
            "        page = browser.new_page()\n\n"
        )

        script_footer = (
            "        time.sleep(3) # Keep browser open for 3 seconds to see the result\n"
            "        browser.close()\n\n"
            "if __name__ == '__main__':\n"
            "    run_automation()\n"
        )

        script_body = []

        for i, action_str in enumerate(actions_list):
            code = self._parse_action_to_code(action_str)

            indented_code = "        " + code.replace("\n", "\n        ")

            script_body.append(indented_code)
            script_body.append("")  # Empty line for separation

        final_script = script_header + "\n".join(script_body) + script_footer

        try:
            with open(self.output_path, "w") as f:
                f.write(final_script)
        except Exception as e:
            print(f"Error writing script to file: {e}")



import math
from collections import Counter
from typing import List
from urllib.parse import urlparse

from playwright.async_api import Page

from pyba.utils.structure import CleanedDOM


def url_entropy(url) -> int:
    """
    Computes the shannon entropy of a URL useful for determining which URLs to
    keep during the general DOM href extraction
    """
    counts = Counter(url)
    total = len(url)
    return -sum((count / total) * math.log2(count / total) for count in counts.values())


def is_absolute_url(url: str) -> bool:
    """
    Determines if a URL is absolute or relative. Used in fixing relative URLs
    in case of goto actions in playwright
    """
    parsed = urlparse(url)
    return bool(parsed.scheme and parsed.netloc)


async def initial_page_setup(page: Page) -> CleanedDOM:
    """
    Helper function for main: goto for the initial page -> Optimisation
    """
    start_page = "https://search.brave.com"

    await page.goto(start_page)

    cleaned_dom = CleanedDOM(
        hyperlinks=[],
        input_fields=["#searchbox"],
        clickable_fields=[],
        actual_text=None,
        current_url=start_page,
    )

    return cleaned_dom


def verify_login_page(page_url: str, url_list: List[str]):
    """
    Helper function called inside login engines

    Args:
        `page_url`: The page URL to be checked against a known list
        `url_list`: The know URL list for login sites for the specific website

    Returns:
        bool: Depending on whether this page is one of the login ones or not

    Note: This assumes that all the urls in the `url_list` are ending with a "/".
    """
    parsed = urlparse(page_url)
    normalized_url = f"{parsed.scheme}://{parsed.netloc}{parsed.path}"

    if not normalized_url.endswith("/"):
        normalized_url += "/"

    # Keeping it simple with this right now, later we can make this better
    if normalized_url in url_list:
        return True
    else:
        return False



from pyba.utils.exceptions import ServiceNotSelected, ServerLocationUndefined
from pyba.utils.load_yaml import load_config

config = load_config("general")["main_engine_configs"]


class Provider:
    """
    Class to handle the provider instances.
    """

    def __init__(
        self,
        openai_api_key: str = None,
        gemini_api_key: str = None,
        vertexai_project_id: str = None,
        vertexai_server_location: str = None,
        logger=None,
    ):
        """
        Args:
                openai_api_key: API key for OpenAI models should you want to use that
                vertexai_project_id: Create a VertexAI project to use that instead of OpenAI
                vertexai_server_location: VertexAI server location
                logger: The logger instance
        """

        self.provider: str | None = None
        self.model: str | None = None
        self.openai_api_key: str | None = openai_api_key
        self.vertexai_project_id: str | None = vertexai_project_id
        self.gemini_api_key: str | None = gemini_api_key
        self.location: str | None = vertexai_server_location

        self.log = logger

        self.handle_keys()

    def handle_keys(self):
        if (
            self.openai_api_key is None
            and self.vertexai_project_id is None
            and self.gemini_api_key is None
        ):
            raise ServiceNotSelected()

        if self.vertexai_project_id and self.location is None:
            raise ServerLocationUndefined(self.location)

        if (
            self.openai_api_key
            and self.vertexai_project_id
            or (self.vertexai_project_id and self.gemini_api_key)
            or (self.openai_api_key and self.gemini_api_key)
        ):
            self.log.warning(
                "You've defined more than one LLM keys, we're choosing to go with openai!"
            )
            self.provider = config["openai"]["provider"]
            self.model = config["openai"]["model"]
            self.vertexai_project_id = None
            self.location = None

        elif self.vertexai_project_id:
            # Assuming that we don't have an openai_api_key
            self.provider = config["vertexai"]["provider"]
            self.model = config["vertexai"]["model"]
        elif self.openai_api_key:
            self.provider = config["openai"]["provider"]
            self.model = config["openai"]["model"]
        else:
            self.provider = config["gemini"]["provider"]
            self.model = config["gemini"]["model"]



import asyncio
import uuid
from typing import List, Union

from playwright.async_api import async_playwright
from playwright_stealth import Stealth

from pyba.core.agent.playwright_agent import PlaywrightAgent
from pyba.core.lib import HandleDependencies
from pyba.core.lib.action import perform_action
from pyba.core.lib.code_generation import CodeGeneration
from pyba.core.provider import Provider
from pyba.core.scripts import LoginEngine, ExtractionEngines
from pyba.core.tracing import Tracing
from pyba.database import Database, DatabaseFunctions
from pyba.logger import get_logger
from pyba.utils.common import initial_page_setup
from pyba.utils.exceptions import PromptNotPresent, UnknownSiteChosen, DatabaseNotInitialised
from pyba.utils.load_yaml import load_config

config = load_config("general")


class Engine:
    """
    The main entrypoint for browser automation. This engine exposes the main entry point which is the run() method
    """

    def __init__(
        self,
        openai_api_key: str = None,
        vertexai_project_id: str = None,
        vertexai_server_location: str = None,
        gemini_api_key: str = None,
        headless: bool = config["main_engine_configs"]["headless_mode"],
        handle_dependencies: bool = config["main_engine_configs"]["handle_dependencies"],
        use_logger: bool = config["main_engine_configs"]["use_logger"],
        enable_tracing: bool = config["main_engine_configs"]["enable_tracing"],
        trace_save_directory: str = None,
        database: Database = None,
    ):
        """
        Args:
            `openai_api_key`: API key for OpenAI models should you want to use that
            `vertexai_project_id`: Create a VertexAI project to use that instead of OpenAI
            `vertexai_server_location`: VertexAI server location
            `gemini_api_key`: API key for Gemini-2.5-pro native support without VertexAI
            `headless`: Choose if you want to run in the headless mode or not
            `use_logger`: Choose if you want to use the logger (that is enable logging of data)
            `handle_dependencies`: Choose if you want to automatically install dependencies during runtime
            `enable_tracing`: Choose if you want to enable tracing. This will create a .zip file which you can use in traceviewer
            `trace_save_directory`: The directory where you want the .zip file to be saved

            `database`: An instance of the Database class which will define all database specific configs

        Find these default values at `pyba/config.yaml`
        """
        self.session_id = uuid.uuid4().hex
        self.headless_mode = headless
        self.tracing = enable_tracing
        self.trace_save_directory = trace_save_directory

        # Handle database instances using `db_funcs`
        self.database = database
        self.db_funcs = DatabaseFunctions(self.database) if database else None

        # Initialising the loggering depending on whether the use_logger boolean is on
        self.log = get_logger(use_logger=use_logger)

        self.automated_login_engine_classes = []

        selectors = tuple(config["process_config"]["selectors"])
        self.combined_selector = ", ".join(selectors)

        self.handle_dependencies(handle_dependencies)

        provider_instance = Provider(
            openai_api_key=openai_api_key,
            gemini_api_key=gemini_api_key,
            vertexai_project_id=vertexai_project_id,
            vertexai_server_location=vertexai_server_location,
            logger=self.log,
        )

        self.provider = provider_instance.provider
        self.model = provider_instance.model
        self.openai_api_key = provider_instance.openai_api_key
        self.gemini_api_key = provider_instance.gemini_api_key
        self.vertexai_project_id = provider_instance.vertexai_project_id
        self.location = provider_instance.location

        # Defining the playwright agent with the defined configs
        self.playwright_agent = PlaywrightAgent(engine=self)

    @staticmethod
    def handle_dependencies(handle_dependencies: bool):
        if handle_dependencies:
            HandleDependencies.playwright.handle_dependencies()

    async def run(self, prompt: str = None, automated_login_sites: List[str] = None):
        """
        The most basic implementation for the run function

        Args:
            `prompt`: The user's instructions
                Right now we're assuming that the user's prompt is well defined. In later
                versions we'll come up with a fix for that as well.
        """

        if prompt is None:
            raise PromptNotPresent()

        if automated_login_sites is not None:
            assert isinstance(
                automated_login_sites, list
            ), "Make sure the automated_login_sites is a list!"

            for engine in automated_login_sites:
                # Each engine is going to be a name like "instagram"
                if hasattr(LoginEngine, engine):
                    engine_class = getattr(LoginEngine, engine)
                    self.automated_login_engine_classes.append(engine_class)
                else:
                    raise UnknownSiteChosen(LoginEngine.available_engines())

        async with Stealth().use_async(async_playwright()) as p:
            self.browser = await p.chromium.launch(headless=self.headless_mode)

            tracing = Tracing(
                browser_instance=self.browser,
                session_id=self.session_id,
                enable_tracing=self.tracing,
                trace_save_directory=self.trace_save_directory,
            )
            self.trace_dir = tracing.trace_dir

            self.context = await tracing.initialize_context()
            self.page = await self.context.new_page()
            cleaned_dom = await initial_page_setup(self.page)

            for steps in range(0, config["main_engine_configs"]["max_iteration_steps"]):
                # First check if we need to login and run the scripts
                login_attempted_successfully = False

                # If loginengines have been chosen then self.automated_login_engine_classes will be populated
                if self.automated_login_engine_classes:
                    for engine in self.automated_login_engine_classes:
                        engine_instance = engine(self.page)
                        self.log.info(f"Testing for {engine_instance.engine_name} login engine")
                        # Instead of just running it and checking inside, we can have a simple lookup list
                        out_flag = await engine_instance.run()
                        if out_flag:
                            # This means it was True and we successfully logged in
                            self.log.success(
                                f"Logged in successfully through the {self.page.url} link"
                            )
                            login_attempted_successfully = True
                            break
                        elif out_flag is None:
                            # This means it wasn't for a login page for this engine
                            pass
                        else:
                            # This means it failed
                            self.log.warning(f"Login attempted at {self.page.url} but failed!")
                if login_attempted_successfully:
                    # Clean the automated_login_engine_classes
                    self.automated_login_engine_classes = None
                    # Update the DOM after a login
                    try:
                        await self.page.wait_for_load_state("networkidle", timeout=2000)
                    except Exception:
                        await asyncio.sleep(2)

                    page_html = await self.page.content()
                    body_text = await self.page.inner_text("body")
                    elements = await self.page.query_selector_all(self.combined_selector)
                    base_url = self.page.url

                    extraction_engine = ExtractionEngines(
                        html=page_html,
                        body_text=body_text,
                        elements=elements,
                        base_url=base_url,
                        page=self.page,
                    )
                    cleaned_dom = await extraction_engine.extract_all()
                    cleaned_dom.current_url = base_url

                    # Jump to the next iteration of the `for` loop
                    continue

                # Say we're going to run only 10 steps so far, so after this no more automation
                # Get an actionable PlaywrightResponse from the models
                try:
                    action = self.playwright_agent.process_action(
                        cleaned_dom=cleaned_dom.to_dict(), user_prompt=prompt
                    )
                except Exception as e:
                    self.log.error(f"something went wrong in obtaining the response: {e}")
                    action = None

                if action is None or all(value is None for value in vars(action).values()):
                    self.log.success("Automation completed, agent has returned None")
                    await self.save_trace()
                    await self.shut_down()
                    try:
                        output = self.playwright_agent.get_output(
                            cleaned_dom=cleaned_dom.to_dict(), user_prompt=prompt
                        )
                        self.log.info(f"This is the output given by the model: {output}")
                        return output
                    except Exception:
                        # Usually a resource exhausted error
                        await asyncio.sleep(10)
                        output = self.playwright_agent.get_output(
                            cleaned_dom=cleaned_dom.to_dict(), user_prompt=prompt
                        )
                        self.log.info(f"This is the output given by the model: {output}")
                        return output

                self.log.action(action)
                if self.db_funcs:
                    self.db_funcs.push_to_episodic_memory(
                        session_id=self.session_id, action=str(action), page_url=str(self.page.url)
                    )
                # If its not None, then perform it
                await perform_action(self.page, action)

                try:
                    await self.page.wait_for_load_state(
                        "networkidle", timeout=1000
                    )  # Wait for a second for network calls to stablize
                    page_html = await self.page.content()
                except Exception:
                    # We might get a "Unable to retrieve content because the page is navigating and changing the content" exception
                    # This might happen because page.content() will start and issue an evaluate, while the page is reloading and making network calls
                    # So, once it gets a response, it commits it and clears the execution contents so page.content() fails.
                    # See https://github.com/microsoft/playwright/issues/16108

                    # We might choose to wait for networkidle -> https://github.com/microsoft/playwright/issues/22897
                    try:
                        await self.page.wait_for_load_state("networkidle", timeout=2000)
                    except Exception:
                        # If networkidle never happens, then we'll try a direct wait
                        await asyncio.sleep(3)

                    page_html = await self.page.content()

                body_text = await self.page.inner_text("body")
                elements = await self.page.query_selector_all(self.combined_selector)
                base_url = self.page.url

                # Then we need to extract the new cleaned_dom from the page
                # Passing in known_fields for the input fields that we already know off so that
                # its easier for the extraction engine to work
                extraction_engine = ExtractionEngines(
                    html=page_html,
                    body_text=body_text,
                    elements=elements,
                    base_url=base_url,
                    page=self.page,
                )

                # Perform an all out extraction
                cleaned_dom = await extraction_engine.extract_all()

                cleaned_dom.current_url = base_url

        await self.save_trace()
        await self.shut_down()

    async def save_trace(self):
        """
        Endpoint to save the trace if required
        """
        if self.tracing:
            trace_path = self.trace_dir / f"{self.session_id}_trace.zip"
            self.log.info(f"This is the tracepath: {trace_path}")
            await self.context.tracing.stop(path=str(trace_path))

    async def shut_down(self):
        """
        Function to cleanly close the existing browsers and contexts. This also saves
        the traces in the provided trace_dir by the user or the default.
        """
        await self.context.close()
        await self.browser.close()

    def sync_run(
        self, prompt: str = None, automated_login_sites: List[str] = None
    ) -> Union[str, None]:
        """
        Sync endpoint for running the above function
        """
        output = asyncio.run(self.run(prompt=prompt, automated_login_sites=automated_login_sites))

        if output:
            return output

    def generate_code(self, output_path: str) -> bool:
        """
        Function end-point for code generation

        Args:
            `output_path`: output file path to save the generated code to
        """
        if not self.db_funcs:
            raise DatabaseNotInitialised()

        codegen = CodeGeneration(
            session_id=self.session_id, output_path=output_path, database_funcs=self.db_funcs
        )
        codegen.generate_script()
        self.log.info(f"Created the script at: {output_path}")
        return True



from typing import Tuple, Dict

# VertexAI and gemini
from google import genai
from google.genai.types import GenerateContentConfig

# OpenAI
from openai import OpenAI

from pyba.utils.load_yaml import load_config
from pyba.utils.prompts import system_instruction, output_system_instruction
from pyba.utils.structure import PlaywrightResponse, OutputResponseFormat

config = load_config("general")


class LLMFactory:
    """
    Class for handling different types of LLM. The supported LLMs are:

    1. OpenAI - GPT-4o, GPT-3.5-turbo
    2. VertexAI - Gemini-2.5-pro
    3. Native gemini-2.5-pro API
    """

    def __init__(self, engine):
        """
        Initialise the engine parameters as given by the user

        Args;
                engine: The LLM parameters provided by the user
        """
        self.engine = engine
        self.vertexai_client = None
        self.openai_client = None
        self.gemini_client = None

        if self.engine.provider == "openai":
            self.openai_client = self._initialize_openai_client()
        elif self.engine.provider == "vertexai":
            self.vertexai_client = self._initialize_vertexai_client()
        else:
            self.gemini_client = self._initialize_gemini_client()

    def _initialize_vertexai_client(self):
        """
        Initialises the VertexAI client using engine parameters
        """

        vertexai_client = genai.Client(
            vertexai=True, project=self.engine.vertexai_project_id, location=self.engine.location
        )

        return vertexai_client

    def _initialize_vertexai_agent(self, system_instruction: str, response_schema):
        """
        Initiaises a VertexAI agent

        Args:
                `system_instruction`: The system instruction for the agent
                `response_schema`: The response schema for the Agent
        """
        assert system_instruction is not None and response_schema is not None

        agent = self.vertexai_client.chats.create(
            model=self.engine.model,
            config=GenerateContentConfig(
                temperature=0,
                system_instruction=system_instruction,
                response_schema=response_schema,
                response_mime_type="application/json",
            ),
        )

        return agent

    def _initialize_openai_client(self):
        """
        Initialize the OpenAI client using engine parameters
        """
        openai_client = OpenAI(api_key=self.engine.openai_api_key)
        return openai_client

    def _initialize_openai_agent(self, system_instruction: str, response_schema) -> Dict:
        """
        Initialize the OpenAI agent

        Args:
                `system_instruction`: The system instruction for the agent
                `response_schema`: The response type for the agent

        Returns:
                Dictionary of the agent parameters
        """

        agent = {
            "client": self.openai_client,
            "system_instruction": system_instruction,
            "model": config["main_engine_configs"]["openai"]["model"],
            "response_format": response_schema,
        }

        return agent

    def _initialize_gemini_client(self):
        """
        Initialises the native gemini-2.5-pro client (without VertexAI)
        """
        gemini_client = genai.Client(vertexai=False, api_key=self.engine.gemini_api_key)
        return gemini_client

    def _initialize_gemini_agent(self, system_instruction: str, response_schema) -> Dict:
        """
        Initilse the Gemini Agent

        Args:
            `system_instruction`: The system instruction for the agent
            `response_schema`: The response type for the agent

        Returns:
            Dictionary of the agent parameters
        """
        agent = {
            "client": self.gemini_client,
            "system_instruction": system_instruction,
            "model": self.engine.model,
            "response_format": response_schema,
        }

        return agent

    def create_agentic_pair(self, init_method) -> Tuple:
        """
        Create the action and output agents for different LLMs

        Args:
            `init_method`: Function to initialise the respective LLM agent

        Returns:
            A tuple containing the action and output agent
        """

        action_agent = init_method(
            system_instruction=system_instruction, response_schema=PlaywrightResponse
        )
        output_agent = init_method(
            system_instruction=output_system_instruction, response_schema=OutputResponseFormat
        )

        return (action_agent, output_agent)

    def get_agent(self) -> Tuple:
        """
        Endpoint to return the agents depending on the LLM called for

        Returns:
                A tuple containing the main agent and the output agent for a particular provider
        """
        if self.engine.provider == "openai":
            agents = self.create_agentic_pair(self._initialize_openai_agent)
        elif self.engine.provider == "vertexai":
            agents = self.create_agentic_pair(self._initialize_vertexai_agent)
        else:
            agents = self.create_agentic_pair(self._initialize_gemini_agent)

        return agents



import os
import re
import subprocess
import sys
from pathlib import Path


class PlaywrightDependencies:
    """
    We'll make use of this class if the user wants us to handle the dependencies. Usually, this
    shouldn't be the case because its like 2 commands to manage all playwright deps.
    """

    @staticmethod
    def _get_playwright_browsers_path() -> Path:
        """
        Determines the default or configured path where Playwright browsers are
        installed. An OS agnostic check for determining installed playwright browsers.

        Uses the following paths:
            `windows`: "AppData/Local/ms-playwright"
            `macOS`: "~/Library/Caches/ms-playwright"
            `linux`: ".cache/ms-playwright"

        Note: OSes like freebsd haven't been checked
        """
        env_path = os.environ.get("PLAYWRIGHT_BROWSERS_PATH")
        if env_path:
            return Path(env_path).expanduser().resolve()

        if os.name == "nt":
            base = (
                Path(os.environ.get("LOCALAPPDATA", Path.home() / "AppData" / "Local"))
                / "ms-playwright"
            )
        elif sys.platform == "darwin":
            base = Path.home() / "Library" / "Caches" / "ms-playwright"
        else:
            # This is buggy
            base = Path.home() / ".cache" / "ms-playwright"
        return base

    @staticmethod
    def _get_expected_browser_dirs() -> list[str]:
        """
        Runs 'playwright install --dry-run --json' to determine which browser
        directories should exist, based on the current Playwright version.
        """
        try:
            result = subprocess.run(
                [sys.executable, "-m", "playwright", "install", "--dry-run"],
                capture_output=True,
                text=True,
                check=False,
            )

            if result.returncode != 0 or not result.stdout.strip():
                return []

            data = result.stdout
            matches = re.findall(r"Install location:\s+(.+)", data)
            expected_dirs = [Path(path.strip()).name for path in matches if path.strip()]
            return expected_dirs
        except Exception as e:
            print(f"Error determining expected browser directories: {e}")
            return []

    @staticmethod
    def check_playwright_browsers_installed() -> bool:
        """
        Checks if all expected Playwright browsers are installed in the cache directory.
        """
        try:
            base_path = PlaywrightDependencies._get_playwright_browsers_path()
            expected_dirs = PlaywrightDependencies._get_expected_browser_dirs()

            if not expected_dirs:
                return False

            # Get the names of existing directories inside the Playwright cache path
            existing_dirs = {p.name for p in base_path.glob("*") if p.is_dir()}
            missing_dirs = [name for name in expected_dirs if name not in existing_dirs]

            return len(missing_dirs) == 0

        except Exception as e:
            print(f"An unexpected error occurred during browser check: {e}", file=sys.stderr)
            return False

    @staticmethod
    def install_playwright_browsers():
        """
        Install Playwright browsers automatically.
        """
        print("Installing Playwright browsers...")
        subprocess.run([sys.executable, "-m", "playwright", "install"], check=True)

    @staticmethod
    def check_missing_dependencies():
        """
        Run a test to identify missing dependencies using `playwright install-deps --dry-run`.
        """
        try:
            result = subprocess.run(
                [sys.executable, "-m", "playwright", "install-deps", "--dry-run"],
                capture_output=True,
                text=True,
                check=False,
            )
            output = result.stdout.strip() + "\n" + result.stderr.strip()
            # Filter only the lines that mention missing libraries
            missing_lines = [
                line
                for line in output.splitlines()
                if "Missing" in line or line.strip().endswith(".so") or "║" in line
            ]
            if missing_lines:
                print("The following dependencies are missing: ")
                print("\n".join(missing_lines))
                print(
                    "Please install them manually using your package manager or run: playwright install-deps"
                )

            else:
                print("All playwright dependencies present")
        except FileNotFoundError:
            print("Could not run playwright. Is it installed in this environment?")
        except Exception as e:
            print(f"An error occurred during dependency check: {e}")

    @staticmethod
    def handle_dependencies():
        # Step 1: Check if browsers are installed
        if PlaywrightDependencies.check_playwright_browsers_installed():
            print("Playwright browsers are already installed")
        else:
            print("Playwright browsers not found.")
            choice = input("Do you want to install them automatically? (y/n): ").strip().lower()
            if choice == "y":
                try:
                    PlaywrightDependencies.install_playwright_browsers()
                    print("Installation complete.")
                except subprocess.CalledProcessError:
                    print("Browser installation failed. Check the output above for errors.")
            else:
                print("Please install browsers using: playwright install")
        # Step 2: Check missing system dependencies
        PlaywrightDependencies.check_missing_dependencies()


class HandleDependencies:
    playwright = PlaywrightDependencies



import json
import time
from typing import Optional

from pyba.database.database import Database
from pyba.database.models import EpisodicMemory


class DatabaseFunctions:
    """
    Composition class for the database functions
    """

    def __init__(self, database: Database):
        """
        Args:
            `database`: The database instance for commiting

        If database is none, it doesn't initialise the database session
        """
        if database is None:
            return
        self.database = database
        self.session = self.database.session

    def submit_query_with_retry(self):
        """
        Function to send submit based queries to db
        (such as insert and update or delete), it retries 100 times if
        connection returned an error.

        Args:
            `session`: session to commit

        Returns:
            True if submitted success otherwise False
        """
        if not hasattr(self, "session"):
            return False

        try:
            for _ in range(1, 100):
                try:
                    self.session.commit()
                    return True
                except Exception:
                    time.sleep(0.1)
        except Exception:
            self.session.rollback()
            return False

        self.session.rollback()
        return False

    def push_to_episodic_memory(self, session_id: str, action: str, page_url: str) -> bool:
        """
        Pushes a new action and page_url onto the stack for a given session_id.
        It retrieves the existing record, appends the new values as JSON strings,
        and updates/inserts the record.

        Args:
            session_id: The unique session ID.
            action: The action string to be pushed.
            page_url: The page URL string to be pushed.

        Returns:
            True if the operation was successful, otherwise False.
        """
        if not hasattr(self, "session"):
            return False
        try:
            memory_record = (
                self.session.query(EpisodicMemory)
                .filter(EpisodicMemory.session_id == session_id)
                .one_or_none()
            )

            if memory_record:
                try:
                    actions_list = json.loads(memory_record.actions)
                    page_url_list = json.loads(memory_record.page_url)
                except json.JSONDecodeError:
                    # If stored data is not a valid json, refresh it with a new list
                    actions_list = []
                    page_url_list = []

                actions_list.append(action)
                page_url_list.append(page_url)

                memory_record.actions = json.dumps(actions_list)
                memory_record.page_url = json.dumps(page_url_list)

            else:
                new_memory = EpisodicMemory(
                    session_id=session_id,
                    actions=json.dumps([action]),
                    page_url=json.dumps([page_url]),
                )
                self.session.add(new_memory)

            return self.submit_query_with_retry()

        except Exception:
            self.session.rollback()
            return False
        finally:
            self.session.close()

    def get_episodic_memory_by_session_id(self, session_id: str) -> Optional[EpisodicMemory]:
        """
        Retrieves an episodic memory record by its `session_id`.

        Args:
            `session_id`: The unique session ID to query for.
        Returns:
            An EpisodicMemory object if found, else None.
        """
        # Check if session exists
        if not hasattr(self, "session"):
            return None
        try:
            memory = self.session.get(EpisodicMemory, session_id)
            return memory
        except Exception:
            return None



general_prompt = """

You are the Brain of a browser automation engine.

Your goal is to interpret the user's intent and decide the next Playwright actions that move toward completing the task. You are currently viewing a snapshot of the webpage's DOM, represented as structured information.

---

### USER GOAL
{user_prompt}

---

### CURRENT PAGE CONTEXT (Cleaned DOM)

# Note that we open by default at Google.com, so the first page will be google's landing page.

**Hyperlinks (clickable anchors or navigation targets):**
{hyperlinks}

**Input Fields (fillable text boxes or form elements):**
{input_fields}

**Clickable Fields (buttons, divs, spans, or elements with onClick handlers):**
{clickable_fields}

**Visible Text (actual text content present on the page):**
{actual_text}

**Current page URL**
{current_url}

---
### YOUR JOB
Using the above DOM context and the user's goal:

1. Understand what the **next single Playwright action** should be to move closer to the goal.
2. You must produce **exactly one atomic PlaywrightAction** per step.
3. **Only one field** of the `PlaywrightAction` schema (apart from required pairs like `fill_selector` + `fill_value`, or `press_selector` + `press_key`) should be non-null at any time.
4. All other fields in that action must be `null` (or absent from the JSON).
5. Do not combine multiple operations in a single action. For example, if you need to fill an input and then press Enter, these must happen in **two separate sequential steps**.
6. Choose only selectors that exist in the DOM snapshot provided.
7. Keep the plan minimal, sequential, and reliable.
8. If you recently filled an input relevant to the user’s goal, the next step will likely be to press Enter on that same field to submit it.
9. If no clickable or fillable elements match the goal, pick the most relevant visible input field (based on user intent) and press Enter on it.

---

### CRITICAL CONSTRAINT
At any given time, only **one actionable field** from the schema should be active (non-null).  
Every other field should be `None` or omitted.

For example:

- Allowed:
Step1:

{{
  "actions": [
    {{	
	    "fill_selector": "input[name='q']",
	   	"fill_value": "Python"
    }}
  ]
}}

Step2:
{{
  "actions": [
    {{"press_selector": "input[name='q']"}}
  ]
}}

step3:
{{
  "actions": [
    {{"press_key": "Enter"}}
  ]
}}


- Not allowed:
{{
  "actions": [
    {{ "fill_selector": "input[name='q']", "fill_value": "Python", "press_selector": "input[name='q']", "press_key": "Enter" }}
  ]
}}
---

###  OUTPUT FORMAT
You must output **only a valid JSON object** of type `PlaywrightResponse`.

If you believe the automation has completed and there is nothing more to do, return `None`.

---

### EXAMPLES

**Example 1:**  

User wants to search for “Python Playwright tutorial” on Google.

- You find an input field with placeholder “Search” → Fill it.  
- Next step → Press Enter on that same input.

{{
  "actions": [
    {{
      "fill_selector": "input[name='q']",
      "fill_value": "Python Playwright tutorial"
    }}
  ]
}}

Then, in the following step:
{{
  "actions": [
    {{
      "press_selector": "input[name='q']",
      "press_key": "Enter"
    }}
  ]
}}

Example 2:
User wants to submit a form but no “submit” button is visible.

- Press Enter on the most relevant visible input field that was recently filled or matches the context.

{{
  "actions": [
    {{
      "press_selector": "input[name='email']",
      "press_key": "Enter"
    }}
  ]
}}


IMPORTANT:

YOU CAN ALWAYS WORK WITH THE ASSUMPTION THAT WHATEVER YOU HAD PREVIOUSLY ASKED FOR HAS BEEN ACHIEVED. Look at the available information about the page
and make an inference of what is on the page and what is requested by the user.

"""



from pathlib import Path

from playwright.async_api import Page

from pyba.utils.load_yaml import load_config


class YouTubeDOMExtraction:
    """
    Extracts links along with their texts from a youtube page. This is specifically designed for youtube pages, and can be used
    either when a search result is queried and videos are being browsed or when a video is playing and something else needs to
    be clicked.

    This provides an exhaustive list of the valid selectors and buttons which are needed for interacting on YouTube.
    """

    def __init__(self, page: Page):
        """
        Evaluates javascript inside the browser page

        1. links on the page along with their titles (for all visible videos)
        2. input fields for searches and comments
        3. Buttons for like and dislike etc.
        """
        self.page = page
        self.config = load_config("extraction")["youtube"]

        # TODO: take care of the fact that we might define multiple functions later
        # We'll be using the same javascript for all extraction functions to be executed in the browser
        js_file_path = Path(__file__).parent.parent / "js/extractions.js"
        self.js_function_string = js_file_path.read_text()

    async def extract_links_and_titles(self):
        """
        Extracts all the video links and their title names from a YouTube page. Its simply checking for
        all possible `/watch?v=` type selectors and querying their names. We're first writing the vanilla
        Javascript which is to be executed in the browser session to get all the results from it, otherwise
        we'd need to use BeautifulSoup.
        """

        # TODO: As a fallback mechanism we can also use bs4 in here
        videos = await self.page.evaluate(self.js_function_string, self.config)
        # We don't want to touch the page again or any other system
        return videos

    async def extract(self):
        videos = await self.extract_links_and_titles()
        return videos



import random


class Retry:
    """
    Class to implement the exponential backoff and retry logic. Note that this backoff and retry will be
    blocking for that specific context.

    Defines the following variables:

    `exponential_base`: 2 (we're using base 2)
    `base_timeout`: 1 second
    `max_backoff_time`: 60 seconds
    """

    def __init__(self):
        self.base = 2
        self.base_timeout = 1
        self.max_backoff_time = 60

    def calculate_next_time(self, attempt_number):
        """
        Function to calculate the next wait time in seconds

        Args:
                `attempt_number`: The number of failed attempts
        """

        delay = self.base_timeout * (self.base ** (attempt_number - 1))
        delay = min(delay, self.max_backoff_time)
        jitter = random.uniform(0, delay / 2)
        return delay + jitter



from pathlib import Path

from pyba.utils.load_yaml import load_config

config = load_config("general")["main_engine_configs"]


class Tracing:
    """
    Class to manage all tracing functions
    """

    def __init__(
        self,
        browser_instance,
        session_id: str,
        enable_tracing: bool = False,
        trace_save_directory: str = None,
        screenshots: bool = False,
        snapshots: bool = False,
        sources: bool = False,
    ):
        """
        Args:
                `browser_instance`: The browser instance being used under the main async with statement
                `session_id`: A unique identifier for this session
                `enable_tracing`: A boolean to indicate the use of tracing
                `trace_save_directory`: Directory to save the traces

        """
        self.browser = browser_instance

        self.session_id = session_id
        self.enable_tracing = enable_tracing
        self.trace_save_directory = trace_save_directory

        self.screenshots: bool = config["tracing"]["screenshots"] | screenshots
        self.snapshots: bool = config["tracing"]["snapshots"] | snapshots
        self.sources: bool = config["tracing"]["sources"] | sources

        if self.trace_save_directory is None:
            # This means we revert to default, which is CWD
            trace_save_directory = str(Path.cwd())
        else:
            trace_save_directory = self.trace_save_directory

        self.trace_dir = Path(trace_save_directory)
        self.trace_dir.mkdir(parents=True, exist_ok=True)

        self.har_file_path = self.trace_dir / f"{self.session_id}_network.har"

    async def initialize_context(self):
        if self.enable_tracing:
            context = await self.browser.new_context(
                record_har_path=self.har_file_path,  # HAR file output
                record_har_content=config["tracing"][
                    "record_har_content"
                ],  # include request/response bodies
            )

            await context.tracing.start(
                screenshots=self.screenshots,
                snapshots=self.snapshots,
                sources=self.sources,
            )

        else:
            context = await self.browser.new_context()

        return context



# General prompt to get the output after a successful scan from the model
output_prompt = """

You are the Brain of a browser automation engine.

Your goal is to interpret the user's intent and output a string that will complete the user's goal.
You are currently viewing a snapshot of the webpage's DOM, represented as structured information.

---

### USER GOAL
{user_prompt}

---

### CURRENT PAGE CONTEXT (Cleaned DOM)

**Hyperlinks (clickable anchors or navigation targets):**
{hyperlinks}

**Visible Text (actual text content present on the page):**
{actual_text}

**Current page URL**
{current_url}

---
### YOUR JOB

You need to output a string that will achieve the user's goal. If the user has requested for no such output, then return a simple None.
"""



import json
import time
from types import SimpleNamespace
from typing import Dict, List, Union, Any

from pyba.core.agent.llm_factory import LLMFactory
from pyba.utils.prompts import general_prompt, output_prompt
from pyba.utils.retry import Retry
from pyba.utils.structure import PlaywrightResponse


class PlaywrightAgent(Retry):
    """
    Defines the playwright agent's actions

    Provides two endpoints:
        - `process_action`: for returning the right action on a page
        - `get_output`: for summarizing the chat and returning a string
    """

    def __init__(self, engine) -> None:
        """
        Args:
            `engine`: holds all the arguments from the user

        Initialises the agents using the `.get_agent()` entrypoint from the LLMFactory
        """
        super().__init__()  # Initiaising the retry variables
        self.attempt_number = 1
        self.engine = engine
        self.llm_factory = LLMFactory(engine=self.engine)

        self.action_agent, self.output_agent = self.llm_factory.get_agent()

    def _initialise_prompt(
        self, cleaned_dom: Dict[str, Union[List, str]], user_prompt: str, main_instruction: str
    ):
        """
        Method to initailise the main instruction for any agent

        Args:
            `cleaned_dom`: A dictionary containing nicely formatted DOM elements
            `user_prompt`: The instructions given by the user
            `main_instruction`: The prompt for the playwright agent
        """

        # Adding the user_prompt to the DOM to make it easier to format the prompt
        cleaned_dom["user_prompt"] = user_prompt
        prompt = main_instruction.format(**cleaned_dom)

        return prompt

    def _initialise_openai_arguments(
        self, system_instruction: str, prompt: str, model_name: str
    ) -> Dict[str, List[Dict[str, str]]]:
        """
        Initialises the arguments for OpenAI agents

        Args:
            `system_instruction`: The system instruction for the agent
            `prompt`: The current prompt for the agent
            `model_name`: The OpenAI model name

        Returns:
            An arguments dictionary which can be directly passed to OpenAI agents
        """

        messages = [
            {"role": "system", "content": system_instruction},
            {"role": "user", "content": prompt},
        ]

        kwargs = {
            "model": model_name,
            "messages": messages,
        }

        return kwargs

    def _call_model(self, agent: Any, prompt: str, agent_type: str) -> Any:
        """
        Generic method to call the correct LLM provider and parse the response.

        Args:
            agent: The agent to use (action_agent or output_agent)
            prompt: The fully formatted prompt string
            agent_type: "action" or "output", to determine parsing logic

        Returns:
            The parsed response (SimpleNamespace for action, str for output)

        Uses the attempt_number to give ou
        """
        if self.engine.provider == "openai":
            arguments = self._initialise_openai_arguments(
                system_instruction=agent["system_instruction"],
                prompt=prompt,
                model_name=agent["model"],
            )

            while True:
                try:
                    response = agent["client"].chat.completions.parse(
                        **arguments, response_format=agent["response_format"]
                    )
                    self.attempt_number = 1
                    break
                except Exception:
                    # If we hit a rate limit, calculate the time to wait and retry
                    wait_time = self.calculate_next_time(self.attempt_number)
                    time.sleep(wait_time)  # wait_time is in seconds
                    self.attempt_number += 1

            parsed_json = json.loads(response.choices[0].message.content)

            # Parse based on agent type
            if agent_type == "action":
                return SimpleNamespace(**parsed_json.get("actions")[0])
            elif agent_type == "output":
                return str(parsed_json.get("output"))

        elif self.engine.provider == "vertexai":  # VertexAI logic
            while True:
                try:
                    response = agent.send_message(prompt)
                    self.attempt_number = 1
                    break
                except Exception:
                    wait_time = self.calculate_next_time(self.attempt_number)
                    time.sleep(wait_time)
                    self.attempt_number += 1

            try:
                parsed_object = getattr(
                    response, "output_parsed", getattr(response, "parsed", None)
                )

                if not parsed_object:
                    print("No parsed object found in VertexAI response.")
                    return None

                # Parse based on agent type
                if agent_type == "action":
                    if hasattr(parsed_object, "actions") and parsed_object.actions:
                        return parsed_object.actions[0]
                    raise IndexError("No 'actions' found in VertexAI response.")
                elif agent_type == "output":
                    if hasattr(parsed_object, "output") and parsed_object.output:
                        return str(parsed_object.output)
                    raise IndexError("No 'output' found in VertexAI response.")

            except Exception as e:
                print(f"Unable to parse the output from VertexAI response: {e}")

        else:  # Using gemini
            gemini_config = {
                "response_mime_type": "application/json",
                "response_json_schema": agent["response_format"].model_json_schema(),
                "system_instruction": agent["system_instruction"],
            }

            while True:
                try:
                    response = agent["client"].models.generate_content(
                        model=agent["model"],
                        contents=prompt,
                        config=gemini_config,
                    )
                    self.attempt_number = 1
                    break
                except Exception:
                    wait_time = self.calculate_next_time(self.attempt_number)
                    time.sleep(wait_time)
                    self.attempt_number += 1

            action = agent["response_format"].model_validate_json(response.text)
            return action.actions[0]

    def process_action(
        self, cleaned_dom: Dict[str, Union[List, str]], user_prompt: str
    ) -> PlaywrightResponse:
        """
        Method to process the DOM and provide an actionable playwright response

        Args:
            `cleaned_dom`: Dictionary of the extracted items from the DOM
                - `hyperlinks`: List
                - `input_fields` (basically all fillable boxes): List
                - `clickable_fields`: List
                - `actual_text`: string
            `user_prompt`: The instructions given by the user

            We're assuming this to be well explained. In later versions we'll
            add one more layer on top for plan generation and better commands

            output: A predefined pydantic model
        """
        prompt = self._initialise_prompt(
            cleaned_dom=cleaned_dom, user_prompt=user_prompt, main_instruction=general_prompt
        )

        return self._call_model(agent=self.action_agent, prompt=prompt, agent_type="action")

    def get_output(self, cleaned_dom: Dict[str, Union[List, str]], user_prompt: str) -> str:
        """
        Method to get the final output from the model if the user requested for one
        """

        prompt = self._initialise_prompt(
            cleaned_dom=cleaned_dom, user_prompt=user_prompt, main_instruction=output_prompt
        )

        return self._call_model(agent=self.output_agent, prompt=prompt, agent_type="output")



import asyncio
import os
import urllib.parse
from abc import ABC, abstractmethod
from typing import Optional

from dotenv import load_dotenv
from playwright.async_api import Page

from pyba.utils.common import verify_login_page
from pyba.utils.exceptions import CredentialsnotSpecified
from pyba.utils.load_yaml import load_config

load_dotenv()  # Loading the username and passwords


class BaseLogin(ABC):
    """
    The base class for all login engines.
    This handles common logic like credential loading, page verification,
    and 2FA waiting.
    """

    def __init__(self, page: Page, engine_name: str) -> None:
        self.page = page
        self.engine_name = engine_name

        self.config = load_config("general")["automated_login_configs"][self.engine_name]
        self.username = os.getenv(f"{self.engine_name}_username")
        self.password = os.getenv(f"{self.engine_name}_password")

        if self.username is None or self.password is None:
            raise CredentialsnotSpecified(self.engine_name)

        self.uses_2FA = self.config["uses_2FA"]
        self.final_2FA_url = self.config["2FA_wait_value"]

    @abstractmethod
    async def _perform_login(self) -> bool:
        """
        This will be implemented in the main LoginEngines
        It should return True on success and False on failure.
        """
        raise NotImplementedError

    async def _handle_2fa(self) -> None:
        """
        Blocking wait until the user completes the 2FA step
        by polling the current URL.
        """
        while True:
            current_url = self.page.url
            hostname = urllib.parse.urlparse(current_url).hostname or ""

            if hostname.endswith(
                self.final_2FA_url
            ):  # Only when we reach the required domain name, we'll break
                break

            # Continous polling, not the best way but works for now
            await asyncio.sleep(1)

    async def run(self) -> Optional[bool]:
        """
        The main login execution flow.

        Returns:
            `None` if we're not supposed to launch the automated login script here
            `True/False` if the login was successful or a failure
        """
        val = verify_login_page(page_url=self.page.url, url_list=list(self.config["urls"]))
        if not val:
            return None

        # Run the site-specific login script
        login_successful = await self._perform_login()
        if not login_successful:
            return False

        try:
            await self.page.wait_for_load_state("networkidle", timeout=10000)
        except Exception:
            # It's fine, we'll assume that the login worked nicely
            pass

        if self.uses_2FA:
            await self._handle_2fa()

        return True



from playwright.async_api import Page

from pyba.core.scripts.login.base import BaseLogin


class FacebookLogin(BaseLogin):
    """
    The facebook login engine, inherited form the BaseLogin class
    """

    def __init__(self, page: Page) -> None:
        super().__init__(page, engine_name="facebook")

    async def _perform_login(self) -> bool:
        try:
            await self.page.wait_for_selector(self.config["username_selector"])
            await self.page.fill(self.config["username_selector"], self.username)
            await self.page.fill(self.config["password_selector"], self.password)
            await self.page.click(self.config["submit_selector"])
        except Exception:
            # Now this is bad
            return False

        return True



from typing import Literal

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from pyba.database.mysql import MySQLHandler
from pyba.database.postgres import PostgresHandler
from pyba.database.sqlite import SQLiteHandler
from pyba.utils.load_yaml import load_config

config = load_config("general")["database"]


class Database:
    """
    Client-side database function -> Minimizes the config use
    """

    def __init__(
        self,
        engine: Literal[
            "sqlite", "postgres", "mysql"
        ],  # Optional: Can be specified inside the config as well
        name: str = None,
        host: str = None,
        port: int = None,
        username: str = None,
        password: str = None,
        ssl_mode: Literal["disable", "require"] = None,
    ):
        """
        Args:
                sqlite:
                        `engine`: "sqlite"
                        `name`: path to the database file
                        other details can be left empty

                mysql:
                        `engine`: "mysql"
                        `name`: Name of the mysql database
                        `username` and `password`: For logging into the server
                        `host` and `port`: Location of the server

                        Note: Default port is 3306 for MySQL

                postgres:
                        `engine`: "postgres"
                        `name`: Name of the postgres database
                        `username` and `password`: For logging into the server
                        `host` and `port`: Location of the server

                        Note: Default port is 5432 for MySQL
                        Note: `ssl_mode`: "require" for encrypted databases

        Optionally supports entries defined inside the config as well in case they are not provided here.

        > This engine is the recommended way to define the database structure
        """
        self.engine: str = engine or config["engine"]

        self.name: str = name or config["name"]
        self.host: str = host or config["host"]
        self.port: int = port or config["port"]
        self.username: str = username or config["username"]
        self.password: str = password or config["password"]
        self.ssl_mode: str = ssl_mode or config["ssl_mode"]

        self.database_connection_string = self.build_connection_string(engine_name=self.engine)
        self.session = self.create_connection(engine_name=self.engine)

        self.initialise_tables_and_database()

    def build_connection_string(self, engine_name: Literal["sqlite", "postgres", "mysql"]) -> str:
        """
        Defines connection URLs for the different databases for SQLAlchemy usage

        Args:
                `engine_name`: The database model name for initialisation

        Returns:
                A string for SQLAlchemy connection
        """

        return {
            "postgres": f"postgresql+psycopg2://{self.username}:{self.password}@{self.host}:{self.port}/{self.name}?sslmode={self.ssl_mode}",
            "mysql": f"mysql+pymysql://{self.username}:{self.password}@{self.host}:{self.port}/{self.name}",
            "sqlite": f"sqlite:///{self.name}",
        }[engine_name]

    def create_connection(self, engine_name: Literal["sqlite", "postgres", "mysql"]):
        """
        Function to create connections to the database

        Args:
                `engine_name`: The database engine name

        Returns:
                connection if successful otherwise False
        """
        connection_args = {}

        if engine_name == "sqlite":
            connection_args["check_same_thread"] = False

        try:
            db_engine = create_engine(
                self.database_connection_string,
                connect_args=connection_args,
                pool_size=50,
                pool_pre_ping=True,
            )

            Session = sessionmaker(bind=db_engine)

            return Session()
        except Exception as e:
            # We might get an OperationalError here if the DB doesn't exist yet
            print(f"Couldn't create a connection to the database: {e}")
            return False

    def initialise_tables_and_database(self):
        """
        Method to manage the creation of sqlite, postgres and mysql database and tables
        """
        handler_map = {
            "sqlite": SQLiteHandler,
            "postgres": PostgresHandler,
            "mysql": MySQLHandler,
        }

        HandlerClass = handler_map.get(self.engine)
        handler = HandlerClass(database_engine_configs=self)
        handler.setup()
        print(f"Database setup for {self.engine} complete.")


"""
Example:

from pyba import Engine, Database

database = Database(engine="sqlite", name="/tmp/pyba/pyba.db")
engine = Engine(openai_api_key=os.getenv("openai_api_key"), use_logger=True, handle_dependencies=False, enable_tracing=True, database=database)

output = engine.sync_run(prompt="go to workflow and download my gradecard", automated_login_sites=["workflow_iitm"])

print(f"This is my gradecard details: {output}")

val = engine.save_code(save_path="/tmp/pyba/automation_code.py")

if val:
	print("Automation code saved at /tmp/pyba/automation_code.py")
"""



from dotenv import load_dotenv
from playwright.async_api import Page

from pyba.utils.load_yaml import load_config

# Adjust this import path based on where you save base.py
from .base import BaseLogin

load_dotenv()  # Loading the username and passwords
config = load_config("general")["automated_login_configs"]["instagram"]

# These are specific to instagram's login flow
screen_height = config["click_location"]["default_screen_height"]
x_from_left = config["click_location"]["x_from_left"]
y_from_bottom = config["click_location"]["y_from_bottom"]
y_top_left = screen_height - y_from_bottom


class InstagramLogin(BaseLogin):
    """
    The instagram login engine, inherits from the BaseLogin class. This
    still needs a module level import because it calls for more than what
    is defined in the BaseLogin class.
    """

    def __init__(self, page: Page) -> None:
        super().__init__(page, engine_name="instagram")

    async def _perform_login(self) -> bool:
        try:
            await self.page.wait_for_selector(self.config["username_selector"])
            await self.page.fill(self.config["username_selector"], self.username)
            await self.page.fill(self.config["password_selector"], self.password)
            await self.page.click(self.config["submit_selector"])
        except Exception:
            try:
                # Alternate fields that instagram uses
                await self.page.wait_for_selector(self.config["fallback"]["username_selector"])
                await self.page.fill(self.config["fallback"]["username_selector"], self.username)
                await self.page.fill(self.config["fallback"]["password_selector"], self.password)
                await self.page.click(self.config["submit_selector"])
            except Exception:
                return False

        # There is a not-now button that we need to click
        try:
            await self.page.wait_for_selector(
                self.config["additional_args"]["additional_selector_1"], timeout=30000
            )
            await self.page.click(self.config["additional_args"]["additional_selector_1"])
        except Exception:
            pass

        # Sometimes these things also come up for new updates
        try:
            await self.page.wait_for_selector(
                self.config["additional_args"]["additional_selector_2"], timeout=10000
            )
            await self.page.mouse.click(x_from_left, y_top_left)
        except Exception:
            pass

        return True



from sqlalchemy import Column, Text
from sqlalchemy.orm import declarative_base

Base = declarative_base()


class EpisodicMemory(Base):
    """
    Memory for history logs

    Arguments:
            - `session_id`: A unique session ID for the run
            - `actions`: A JSON string of actions given as output by the model
            - `page_url`: The URL where this action was performed
    """

    __tablename__ = "EpisodicMemory"

    session_id = Column(Text, primary_key=True)
    actions = Column(Text, nullable=False)
    page_url = Column(Text, nullable=False)

    def __repr__(self):
        return ("EpisodicMemory(session_id: {0}, actions: {1}, page_url: {2})").format(
            self.session_id, self.actions, self.page_url
        )



from sqlalchemy import create_engine, text
from sqlalchemy.exc import OperationalError

from pyba.database.models import Base


class PostgresHandler:
    """
    Class handling Postgres database creation
    """

    def __init__(self, database_engine_configs):
        """
        Args:
            `database_engine_configs`: User supplied configurations imported from the Database class
        """

        self.database_engine_configs = database_engine_configs
        self.database_connection_string = self.database_engine_configs.database_connection_string

        self.engine = create_engine(self.database_connection_string)

    def create_postgres_database(self):
        """
        Method to create a postgres database

        Note: If the database is not already created, we need to make one using a different
        connection string: "postgresql+psycopg2://{username}:{password}@{host}:{port}/postgres"

        we connect to the default postgres database and create a new database from there
        """
        try:
            Base.metadata.create_all(self.engine)
        except OperationalError:
            local_engine = create_engine(
                f"postgresql+psycopg2://{self.database_engine_configs.username}:{self.database_engine_configs.password}@{self.database_engine_configs.host}:{self.database_engine_configs.port}/postgres"
            )
            conn = local_engine.connect()
            conn = conn.execution_options(isolation_level="AUTOCOMMIT")

            conn.execute(text(f"CREATE DATABASE {self.database_engine_configs.name}"))
            conn.close()

            Base.metadata.create_all(self.engine)

    def setup(self):
        """
        Entrypoint for database and table creation
        """
        self.create_postgres_database()



from pyba.utils.prompts.system_prompt import system_prompt as system_instruction
from pyba.utils.prompts.general_prompt import general_prompt
from pyba.utils.prompts.output_general_prompt import output_prompt
from pyba.utils.prompts.output_system_prompt import (
    output_system_prompt as output_system_instruction,
)



from sqlalchemy import create_engine, text

from pyba.database.models import Base


class MySQLHandler:
    """
    Class to manage MySQL database and table dependencies
    """

    def __init__(self, database_engine_configs):
        """
        Args:
            `database_engine_configs`: User supplied configurations imported from the Database class
        """
        self.database_engine_configs = database_engine_configs
        self.database_connection_string = self.database_engine_configs.database_connection_string

        self.engine = create_engine(self.database_connection_string)

    def mysql_create_database(self):
        """
        Method to create a MySQL database using the specified credentials
        """

        try:
            with self.engine.connect() as conn:
                existing_databases = conn.execute(text("SHOW DATABASES;"))
                existing_databases = [d[0] for d in existing_databases]

                if self.database_engine_configs.name not in existing_databases:
                    conn.execute(
                        text("CREATE DATABASE {0} ".format(self.database_engine_configs.name))
                    )
        except Exception as e:
            print(e)

    def mysql_create_tables(self):
        """
        Method to create the MySQL tables

        Args:
            None
        Returns:
            True if success otherwise False
        """
        Base.metadata.create_all(self.engine)

    def setup(self):
        """
        Entrypoint for database and table creation
        """
        self.mysql_create_database()
        self.mysql_create_tables()



import sys
from argparse import ArgumentParser

from pyba.version import version


class ArgParser(ArgumentParser):
    """
    Argparser class. This class holds the definitions for each flag
    along with their default values and parses them individually
    """

    def __init__(self) -> None:
        super().__init__(prog="pyba", add_help=True)

        self.add_arguments()
        self.initialise_arguments()

    def add_arguments(self):
        main_options = self.add_argument_group(
            "Main", "The main input options. (as in the 'main' file)"
        )
        main_options.add_argument(
            "-V",
            "--version",
            action="store_true",
            dest="show_version",
            default=False,  # Do not hardcode this, read from a version file
            help="Display the software version",
        )

        base_parser = ArgumentParser(add_help=False)
        base_parser.add_argument(
            "--openai-api-key",
            action="store",
            default=None,
            dest="openai_api_key",
            help="OpenAI API Key if you wish to run the automations using OpenAI's models",
        )

        base_parser.add_argument(
            "--vertexai-project-id",
            action="store",
            default=None,
            dest="vertexai_project_id",
            help="VertexAI project ID if you wish to run the automations using VertexAI's gemini models. Note that vertexai must be properly setup before using this option",
        )

        base_parser.add_argument(
            "--vertexai-server-location",
            action="store",
            default=None,
            dest="vertexai_server_location",
            help="VertexAI server location if you wish to run the automations using VertexAI's gemini models. Note that vertexai must be properly setup before using this option",
        )

        base_parser.add_argument(
            "--gemini-api-key",
            action="store",
            default=None,
            dest="gemini_api_key",
            help="Gemini API key if you wish to run the automations using Gemini-2.5-pro without using VertexAI",
        )

        base_parser.add_argument(
            "--headless",
            action="store_true",
            default=False,
            dest="headless_mode",
            help="Run the automations in headless mode. Useful for web scraping",
        )

        base_parser.add_argument(
            "--handle-deps",
            action="store_true",
            default=False,
            dest="handle_dependencies",
            help="Automatically handle all playwright dependencies",
        )

        base_parser.add_argument(
            "-v",
            action="store_true",
            default=False,
            dest="use_logger",
            help="Start the automations in the verbose mode to get live updates",
        )

        base_parser.add_argument(
            "--enable-tracing",
            action="store_true",
            default=False,
            dest="enable_tracing",
            help="Start's tracing all playwright moves and network requests to create a zip file which can be viewed in playwright traceviewer",
        )

        base_parser.add_argument(
            "--trace-save-dir",
            action="store",
            default=None,
            dest="trace_save_directory",
            help="Directory to save the trace zip file",
        )

        base_parser.add_argument(
            "-t",
            "--task",
            action="store",
            dest="task",
            default=None,
            help="The task which needs to be automated as a string",
        )

        base_parser.add_argument(
            "-L",
            "--login-sites",
            action="append",
            dest="automated_login_sites",
            default=None,
            help="The automated login engine names. Note: To use these you will need to set the username and password in your enviornment. Please read the man page.",
        )

        subparsers = self.add_subparsers(
            title="modes",
            dest="mode",
            required=False,  # Run the base and main flags without setting the mode
            description="Choose the database mode if you wish to log actions and use them for script generation",
            parser_class=ArgumentParser,
        )

        # TODO
        # clarification_mode = subparsers.add_parser("clarify", help="Clarification mode for asking questions to the user about unsure steps", parents=[base_parser])

        # Normal Mode
        subparsers.add_parser(
            "normal",
            help="Does not store logs or ask clarifications during automation",
            parents=[base_parser],
        )

        # Database Mode
        database_mode = subparsers.add_parser(
            "database",
            help="Store logs in the database for script creation",
            parents=[base_parser],
        )
        database_mode.add_argument(
            "-e",
            "--engine",
            action="store",
            default=None,
            required=True,
            dest="database_engine",
            help="The database engine you wish to use",  # This needs to be more explicit
        )

        database_mode.add_argument(
            "-n",
            "--name",
            action="store",
            default=None,
            required=True,
            dest="database_name",
            help="Name of the database in case of MySQL and PostgreSQL, and the path to the database file in case of SQLite",
        )

        database_mode.add_argument(
            "-u",
            "--username",
            action="store",
            default=None,
            dest="database_username",
            help="The username for logging into the database server while using MySQL or PostgreSQL",
        )
        database_mode.add_argument(
            "-p",
            "--password",
            action="store",
            default=None,
            dest="database_password",
            help="The password for logging into the database server while using MySQL or PostgreSQL",
        )
        database_mode.add_argument(
            "-H",
            "--host-name",
            action="store",
            default=None,
            dest="database_host",
            help="The host IP serving the MySQL or PostgreSQL databases",
        )
        database_mode.add_argument(
            "-P",
            "--port",
            action="store",
            default=None,
            dest="database_port",
            help="The host port serving the MySQL or PostgreSQL databases",
        )
        database_mode.add_argument(
            "--ssl-mode",
            action="store",
            default="disabled",
            dest="postgres_ssl_mode",
            help="The ssl_mode for running PostgreSQL databases. Can be disable or required, defaults at disabled",
        )
        database_mode.add_argument(
            "--generate-code",
            action="store_true",
            default=False,
            dest="generate_code",
            help="Use the stored actions to generate an automation script",
        )
        database_mode.add_argument(
            "--code-output-path",
            action="store",
            default=None,
            dest="code_output_path",
            help="The file destination for the output path of the generated code. Defaults at `/tmp/script.py`",
        )

    def initialise_arguments(self):
        """
        check all rules and requirements for ARGS

        Args:
        api_forms: values from nettacker.api

        Returns:
        all ARGS with applied rules
        """
        options = self.parse_args()

        if options.show_version:
            print(f"Software version: {version}")
            sys.exit(0)

        if options.automated_login_sites:
            for site in options.automated_login_sites:
                print(f"Automated login enabled for: {site}")

        # passing all the keys directly to the run function because that handles it using the provider instance
        if options.mode == "database":
            if options.database_engine not in ["sqlite", "mysql", "postgres"]:
                print("Wrong database engine chosen. Please choose from sqlite, mysql or postgres")
                sys.exit(0)

            if not options.code_output_path:
                # Default save to /tmp/pyba_script.py
                options.code_output_path = "/tmp/pyba_script.py"
                print(
                    "Output path not specified, the generated script will be saved at /tmp/pyba_script.py"
                )

        self.arguments = options



from typing import Optional

import colorama
from colorama import Fore, Style

# Initialize colorama to auto-reset styles after each print
colorama.init(autoreset=True)


class Logger:
    """
    Custom logger for PyBA. This only logs if 'use_logger' is enabled.
    """

    def __init__(self, use_logger: bool = False):
        """
        Args:
            use_logger (bool): If False, all logging methods will do nothing.
        """
        self.use_logger = use_logger
        if self.use_logger:
            self.info("Logger initialized. Logging is enabled.")

    def _log(self, prefix: str, message: str, color: str):
        if not self.use_logger:
            return
        print(f"{Style.BRIGHT}{color}{prefix}{Style.NORMAL}{Fore.RESET} {message}")

    def info(self, message: str):
        self._log("[INFO]   ", message, Fore.BLUE)

    def success(self, message: str):
        self._log("[SUCCESS]", message, Fore.GREEN)

    def warning(self, message: str):
        self._log("[WARNING]", message, Fore.YELLOW)

    def error(self, message: str, e: Optional[Exception] = None):
        if e:
            message = f"{message}: {e}"
        self._log("[ERROR]  ", message, Fore.RED)

    def action(self, message: str):
        self._log("[ACTION] ", message, Fore.MAGENTA)


def get_logger(use_logger: bool = False) -> Logger:
    """
    Factory function to get a Logger instance.

    Args:
        use_logger (bool): Flag to enable or disable logging.

    Returns:
        Logger: An instance of the Logger class.
    """
    return Logger(use_logger=use_logger)



from typing import Optional

from pyba import Engine
from pyba.cli.cli_core.arg_parser import ArgParser
from pyba.database import Database


class CLIMain(ArgParser):
    """
    Main class for pyba CLI
    """

    def __init__(self):
        super().__init__()

        self.database = self.initialise_database()
        self.initialise_engine()

        self.generate_code = self.arguments.generate_code
        self.code_output_path = self.arguments.code_output_path

    def initialise_database(self) -> Optional[Database]:
        """
        Helper function to generate and initialise the database
        """
        database = None
        if self.arguments.mode == "database":
            database_configs = {
                "engine": self.arguments.database_engine,
                "name": self.arguments.database_name,
                "username": self.arguments.database_username or None,
                "password": self.arguments.database_password or None,
                "host": self.arguments.database_host or None,
                "port": self.arguments.database_port or None,
                "ssl_mode": self.arguments.postgres_ssl_mode,
            }

            database = Database(**database_configs)

        return database

    def initialise_engine(self):
        """
        Initialises the engine using the provided user parameters
        """

        self.task = self.arguments.task
        self.automated_login_sites = self.arguments.automated_login_sites

        engine_configs = {
            "openai_api_key": self.arguments.openai_api_key,
            "vertexai_project_id": self.arguments.vertexai_project_id,
            "vertexai_server_location": self.arguments.vertexai_server_location,
            "gemini_api_key": self.arguments.gemini_api_key,
            "headless": self.arguments.headless_mode,
            "handle_dependencies": self.arguments.handle_dependencies,
            "use_logger": self.arguments.use_logger,
            "enable_tracing": self.arguments.enable_tracing,
            "trace_save_directory": self.arguments.trace_save_directory,
            "database": self.database,
        }

        self.engine = Engine(**engine_configs)

    def cli_sync_run(self):
        """
        The CLI run function which calls the main runner with the instantiated arguments
        """
        self.engine.sync_run(self.task, automated_login_sites=self.automated_login_sites)
        if self.generate_code:
            self.engine.generate_code(output_path=self.code_output_path)

    def cli_async_run(self):
        """
        The CLI run function for Async endpoints
        """
        self.engine.run(self.task, automated_login_sites=self.automated_login_sites)



from playwright.async_api import Page
from pyba.core.scripts.extractions.general import GeneralDOMExtraction
from pyba.core.scripts.extractions.youtube_ import YouTubeDOMExtraction

import asyncio


class ExtractionEngines:
    """
    Returns all the extraction engines and provides a way to get their names
    """

    general = GeneralDOMExtraction
    youtube = YouTubeDOMExtraction

    @classmethod
    def available_engines(cls):
        return [name for name, value in vars(cls).items() if isinstance(value, type)]

    def __init__(self, html: str, body_text: str, elements: list, base_url: str, page: Page):
        self.html = html
        self.body_text = body_text
        self.elements = elements
        self.base_url = base_url
        self.page = page

        self.output = {}

    async def extract_all(self):
        """
        Create the all encompassing extraction engine
        """
        general = ExtractionEngines.general(
            html=self.html,
            body_text=self.body_text,
            elements=self.elements,
            base_url=self.base_url,
        )
        general_output = await general.extract()
        self.output = general_output

        youtube = ExtractionEngines.youtube(page=self.page)

        if "youtube.com" in self.page.url:
            # Usually the dom extraction is pretty fast but the videos take some time to load up in the javascript
            # Hence a small wait here helps in loading that
            await asyncio.sleep(3)
            youtube_output = await youtube.extract()
            self.output.youtube = youtube_output

        return self.output



# System prompt for the output response

output_system_prompt = """

You are the **Brain** of a browser automation engine.

Your purpose is to reason about a web page’s current state and output a string that will solve the user's query.

---

### What You Receive
You are given:
1. **Task Description (User Goal):**
   - A clear instruction of what needs to be achieved (e.g., "Search for a product and add it to the cart" or "get me the first five links for rickroll on youtube").
   - This represents the next logical intent in the overall browsing flow.

2. **Cleaned DOM (Context of Current Page):**
   A structured dictionary extracted from the web page containing:
   - `hyperlinks`: list of all hyperlink texts or targets
   - `input_fields`: list of all fillable input elements
   - `clickable_fields`: list of clickable elements (buttons, spans, etc.)
   - `actual_text`: visible text content of the page

Use this cleaned DOM to understand what’s visible, available, and interactable on the page right now.

---

### What You Must Output
You must **always output a valid JSON object** of type ``, defined as:

"""



system_prompt = """
You are the **Brain** of a browser automation engine.

Your purpose is to reason about a web page’s current state and produce the **next precise, atomic instruction** for a Playwright automation script.

---

### What You Receive
You are given:
1. **Task Description (User Goal):**
   - A clear instruction of what needs to be achieved (e.g., "Search for a product and add it to the cart").
   - This represents the next logical intent in the overall browsing flow.

2. **Cleaned DOM (Context of Current Page):**
   A structured dictionary extracted from the web page containing:
   - `hyperlinks`: list of all hyperlink texts or targets
   - `input_fields`: list of all fillable input elements
   - `clickable_fields`: list of clickable elements (buttons, spans, etc.)
   - `actual_text`: visible text content of the page

Use this cleaned DOM to understand what’s visible, available, and interactable on the page right now.

---

### What You Must Output
You must **always output a valid JSON object** of type `PlaywrightResponse`, defined as:

```python
class PlaywrightResponse(BaseModel):
    actions: List[PlaywrightAction]
```

Each PlaywrightAction represents one atomic browser action, such as clicking a button, filling a field, typing text, or pressing a key.

### Rules for Output

#### Atomicity:
You must produce exactly one atomic PlaywrightAction per response.

Only one action type (e.g., click, fill_selector, press_key, etc.) may be non-null.

All other fields must be null or absent.

Pairs like (fill_selector + fill_value) or (press_selector + press_key) count as one atomic action.

#### Sequentiality:
Each response represents the next step in the automation.
Complex actions (like filling a field and pressing Enter) must happen over multiple sequential responses, not together.

#### Contextual Validity:
You may only reference elements that exist in the provided cleaned_dom.
Do not invent or assume selectors.

#### Intent Awareness:

Match the user’s goal with what’s currently on screen.

If you just filled an input field relevant to the user goal, the next logical step is often to press Enter on that same field.

If no clickable or fillable element matches the goal, choose the most relevant visible input field (based on goal keywords) and press Enter.

#### Completion Condition:
If the user’s goal appears to be completed or no further action is possible, output None.

### Allowed Examples

Example 1 — Filling a Search Box

{{
  "actions": [
    {{
      "fill_selector": "input[name='q']",
      "fill_value": "Playwright Python"
    }}
  ]
}}


Pressing Enter on the Same Box

{{
  "actions": [
    {{
      "press_selector": "input[name='q']",
      "press_key": "Enter"
    }}
  ]
}}

### Disallowed Examples

Multiple actions in one step:

{{
  "actions": [
    {{
      "fill_selector": "input[name='q']",
      "fill_value": "Playwright",
      "press_selector": "input[name='q']",
      "press_key": "Enter"
    }}
  ]
}}



### Summary

Think like a cautious human tester: one small reliable step at a time.

Never perform two distinct operations in one output.

Always produce structured JSON, never prose or reasoning.

If nothing remains to be done - output None.

"""



from pathlib import Path

import yaml

current_file = Path(__file__)


class ConfigFilePath:
    general = current_file.parent.parent / "config.yaml"
    extraction = current_file.parent.parent / "core/scripts/extractions/extraction_configs.yaml"


def load_config(config_type: str):
    """
    It currently supports two types of config files:

    1. `general` which points to the main config.yaml file
    2. `extraction` which points to the extraction_config.yaml file inside extraction_scripts/
    """
    try:
        config_path = getattr(ConfigFilePath, config_type)
    except AttributeError:
        # This probably will never happen, this is not a userfacing function
        raise ValueError(f"Invalid config type '{config_type}'")

    with open(config_path, "r") as f:
        return yaml.safe_load(f)



import asyncio
import re
from urllib.parse import urljoin

from playwright._impl._errors import Error
from playwright.async_api import Page

from pyba.utils.common import is_absolute_url
from pyba.utils.structure import PlaywrightAction


class PlaywrightActionPerformer:
    """
    The playwright automation class. To add new handles, make a function here
    and define that under perform()

    Below is an exhaustive set of playwright actions that the handler will manage and the dispatcher will execute

    1. Navigation functions
        - handle_navigation
        - handle_back
        - handle_forward
        - handle_reload

    2. Interaction functions
        - handle_input
        - handle_typing
        - handle_click
        - handle_double_click
        - handle_hover
        - handle_checkboxes
        - handle_select
        - handle_file_upload

    3. Keyboard/mouse functions
        - handle_press
        - handle_keyboard_press
        - handle_keyboard_type
        - handle_mouse_move
        - handle_mouse_click

    4. Scrolling
        - handle_scrolling

    5. Waits
        - handle_wait

    6. Javascript functions
        - handle_evaluate_js
        - handle_screenshot
        - handle_download

    7. New pages
        - handle_switch_page
        - handle_new_page
        - handle_close_page

    """

    def __init__(self, page: Page, action: PlaywrightAction):
        self.page = page
        self.action = action

    # -----------------
    # Handle nagivation
    # -----------------
    async def handle_navigation(self):
        """
        Handle's browser naviation -> Opening new websites
        Wait's until the page is loaded
        """
        await self.page.goto(self.action.goto)
        await self.page.wait_for_load_state("domcontentloaded")

    async def handle_back(self):
        await self.page.go_back(wait_until="domcontentloaded")

    async def handle_forward(self):
        await self.page.go_forward(wait_until="domcontentloaded")

    async def handle_reload(self):
        await self.page.reload(wait_until="domcontentloaded")

    # -------------------
    # Handle interactions
    # -------------------
    async def handle_input(self):
        """
        Inputs a value to a selector field
        """
        await self.page.fill(self.action.fill_selector, self.action.fill_value)

    async def handle_typing(self):
        await self.page.type(self.action.type_selector, self.action.type_text)

    async def handle_click(self):
        """
        Handle's clicking elements. Has additional checks to ensure that
        the element is not actually a relational hyperlink.

        This is done in the following ways:

        - We first check if the click element is actually an <a> tag
        - Or if it has a closest ancestory <a> tag

        TODO: MAKE THIS CLEANER

        In either case we extract the href from that <a> tag and directly goto that
        """
        click_target = self.action.click
        if click_target is None:
            return

        # Patch: for absolute URLs
        # If the click string contains a full URL, just extract and goto it.
        url_match = re.search(r"https?://[^\s'\"]+", click_target)
        if url_match:
            href = url_match.group(0)
            await self.page.goto(href)
            await self.page.wait_for_load_state("domcontentloaded")
            return

        locator = self.page.locator(click_target)

        try:
            await locator.scroll_into_view_if_needed()
        except Exception:
            pass

        # Trying to find a hyperlink element (the target or its ancestor)
        # https://playwright.dev/python/docs/api/class-locator#locator-evaluate -> Tests a javascript expression
        try:
            href = await locator.evaluate(
                """
                el => {
                    let link = el.tagName.toLowerCase() === 'a' ? el : el.closest('a');
                    return link ? link.getAttribute('href') : null;
                }
            """
            )
        except Error:
            # Catching a strict mode violation and defaulting to the first click
            # Unfortunately playwright errors aren't fully specific
            first_locator = locator.first
            href = await first_locator.evaluate(
                """
                el => {
                    let link = el.tagName.toLowerCase() === 'a' ? el : el.closest('a');
                    return link ? link.getAttribute('href') : null;
                }
                """
            )
        if href:
            # Handling relative links by checking for a schema and a netloc (host + optional port)
            if not is_absolute_url(href):
                base_url = "/".join(self.page.url.split("/")[0:3])  # This won't be 0:3 always
                href = urljoin(base_url, href)
            await self.page.goto(href)
            await self.page.wait_for_load_state("domcontentloaded")
            return
        else:
            try:
                await locator.click(timeout=5000)
            except Exception:
                await locator.click(force=True, timeout=5000)

    async def handle_double_click(self):
        """
        Handle's double clicking an element
        """
        await self.page.dblclick(self.action.dblclick)

    async def handle_hover(self):
        """
        Handle's hovering over an element to make new actions visible
        """
        await self.page.hover(self.action.hover)

    async def handle_checkboxes(self):
        if self.action.check:
            await self.page.check(self.action.check)
        if self.action.uncheck:
            await self.page.uncheck(self.action.uncheck)

    async def handle_select(self):
        await self.page.select_option(
            self.action.select_selector,
            value=self.action.select_value,
        )

    async def handle_file_upload(self):
        await self.page.set_input_files(
            self.action.upload_selector,
            self.action.upload_path,
        )

    # ----------------------------------
    # Handle mouse/keyboard interactions
    # ----------------------------------
    async def handle_press(self):
        """
        Handles a key press.
        """
        # If a specific selector is provided, press the key on that element.
        if self.action.press_selector and self.action.press_key:
            await self.page.press(self.action.press_selector, self.action.press_key)
        # If no selector is provided, press the key on the entire page.
        elif self.action.press_key:
            await self.page.keyboard.press(self.action.press_key)

    async def handle_keyboard_press(self):
        """
        Handles a keyboard press action on the entire page.
        """
        await self.page.keyboard.press(self.action.keyboard_press)

    async def handle_keyboard_type(self):
        await self.page.keyboard.type(self.action.keyboard_type)

    async def handle_mouse_move(self):
        await self.page.mouse.move(self.action.mouse_move_x or 0, self.action.mouse_move_y or 0)

    async def handle_mouse_click(self):
        await self.page.mouse.click(self.action.mouse_click_x or 0, self.action.mouse_click_y or 0)

    # ----------------
    # Handle scrolling
    # ----------------
    async def handle_scrolling(self):
        """
        Automates manual scrolling (or scrolls to center)
        """
        x = self.action.scroll_x or 0
        y = self.action.scroll_y or 0
        await self.page.mouse.wheel(x, y)

    # ------------
    # Handle waits
    # ------------
    async def handle_wait(self):
        if self.action.wait_selector:
            await self.page.wait_for_selector(
                self.action.wait_selector,
                timeout=self.action.wait_timeout or 5000,
            )
        elif self.action.wait_ms:
            await asyncio.sleep(self.action.wait_ms / 1000)

    # ---------------------------
    # Handle Javascript functions
    # ---------------------------
    async def handle_evaluate_js(self):
        """
        Handles the evaluation of Javascript in the browser enviornement
        and brings the result back to the code

        This is the recommended way to using it.
        ```js
        const href = await page.evaluate(() => document.location.href);
        ```evaluate_js

        We strip the js snippet here for any return statements because those
        aren't required for inline functions.
        """

        if self.action.evaluate_js.startswith("return"):
            self.action.evaluate_js = " ".join(self.action.evaluate_js.split(" ")[1:])
        result = await self.page.evaluate(self.action.evaluate_js)

        # Letting this be here for debugging
        # print("[JS Evaluation Result]:", result)
        return result

    async def handle_screenshot(self):
        await self.page.screenshot(path=self.action.screenshot_path, full_page=True)

    async def handle_download(self):
        async with self.page.expect_download() as download_info:
            await self.page.click(self.action.download_selector)
        download = await download_info.value
        path = await download.path()
        print(f"Downloaded to: {path}")

    # -------------------
    # Handling new pages
    # -------------------
    async def handle_switch_page(self):
        context = self.page.context
        pages = context.pages
        if self.action.switch_page_index is not None and self.action.switch_page_index < len(
            pages
        ):
            self.page = pages[self.action.switch_page_index]
            await self.page.bring_to_front()

    async def handle_new_page(self):
        context = self.page.context
        new_page = await context.new_page()
        await new_page.goto(self.action.new_page)
        await new_page.wait_for_load_state("domcontentloaded")
        self.page = new_page

    async def handle_close_page(self):
        await self.page.close()

    # ----------
    # Dispatcher
    # ----------
    async def perform(self) -> None:
        """
        The main dispatch function.

        All handlers are called here as and when required by the AI models.
        Contains an exhaustive list of all the functions that can be chosen
        during the automation process
        """
        a = self.action

        if a.goto:
            return await self.handle_navigation()
        if a.go_back:
            return await self.handle_back()
        if a.go_forward:
            return await self.handle_forward()
        if a.reload:
            return await self.handle_reload()
        if a.fill_selector and a.fill_value is not None:
            return await self.handle_input()
        if a.type_selector and a.type_text:
            return await self.handle_typing()
        if a.click:
            return await self.handle_click()
        if a.dblclick:
            return await self.handle_double_click()
        if a.hover:
            return await self.handle_hover()
        if a.press_selector or a.press_key:
            return await self.handle_press()
        if a.keyboard_press:
            return await self.handle_keyboard_press()
        if a.keyboard_type:
            return await self.handle_keyboard_type()
        if a.check or a.uncheck:
            return await self.handle_checkboxes()
        if a.select_selector and a.select_value:
            return await self.handle_select()
        if a.upload_selector and a.upload_path:
            return await self.handle_file_upload()
        if a.scroll_x or a.scroll_y:
            return await self.handle_scrolling()
        if a.wait_selector or a.wait_ms:
            return await self.handle_wait()
        if a.evaluate_js:
            return await self.handle_evaluate_js()
        if a.screenshot_path:
            return await self.handle_screenshot()
        if a.download_selector:
            return await self.handle_download()
        if a.new_page:
            return await self.handle_new_page()
        if a.close_page:
            return await self.handle_close_page()
        if a.switch_page_index is not None:
            return await self.handle_switch_page()


async def perform_action(page: Page, action: PlaywrightAction) -> None:
    """
    The entry point function
    """
    # assert isinstance(action, PlaywrightAction), "the input type for action is incorrect!"
    performer = PlaywrightActionPerformer(page, action)
    await performer.perform()



class PromptNotPresent(Exception):
    """
    This exception is raised when the user forgets to enter a prompt to the engine
    """

    def __init__(self):
        super().__init__("Please provide a prompt for us to work on")


class ServiceNotSelected(Exception):
    """
    This exception is raised when the user doesn't set an API key in the engine
    """

    def __init__(self):
        super().__init__("Please set either a VertexAI project ID or an OpenAI key")


class ServerLocationUndefined(Exception):
    """
    This exception is raised when the user doesn't define the server location
    for a VertexAI project.
    """

    def __init__(self, server_location):
        super().__init__(
            f"The server location {server_location} is undefined. Please visit https://cloud.google.com/vertex-ai/docs/general/locations and choose a location that your credits are tied to."
        )


class CredentialsnotSpecified(Exception):
    """
    Exception raised in the login scripts when the relevant credentials haven't been specified
    """

    def __init__(self, site_name: str):
        super().__init__(f"Please specify all the credentials for the {site_name} engine.")


class UnknownSiteChosen(Exception):
    """
    Exception to be raised when the user chooses a site for automated login that isn't implemented yet.
    """

    def __init__(self, sites: list):
        super().__init__(
            f"Unknown site chosen for automated login. The following sites are available: {sites}"
        )


class DatabaseNotInitialised(Exception):
    """
    Exception to be raised when the user asks for automation code generation but has not initialised the database!
    """

    def __init__(self):
        super().__init__(
            "Tried to call for code-generation without logging in a database! Please initialise the database."
        )



# Imports for the extraction scripts
from pyba.core.scripts.extractions import ExtractionEngines

# Imports for the login scripts
from pyba.core.scripts.login import InstagramLogin, FacebookLogin, GmailLogin


class LoginEngine:
    """
    Makes the Automated Login engines available to the main program
    """

    instagram = InstagramLogin
    facebook = FacebookLogin
    gmail = GmailLogin

    @classmethod
    def available_engines(cls):
        return [name for name, value in vars(cls).items() if isinstance(value, type)]



from playwright.async_api import Page

from pyba.core.scripts.login.base import BaseLogin


class GmailLogin(BaseLogin):
    """
    The gmail login engine, inherits from the BaseLogin engine
    """

    def __init__(self, page: Page) -> None:
        super().__init__(page, engine_name="gmail")

    async def _perform_login(self) -> bool:
        try:
            await self.page.wait_for_selector(self.config["username_selector"])
            await self.page.fill(self.config["username_selector"], self.username)
            await self.page.click(self.config["submit_selector"])
        except Exception:
            # Google's too smart
            return False

        try:
            await self.page.wait_for_selector(self.config["password_selector"])
            await self.page.fill(self.config["password_selector"], self.password)
            await self.page.click(self.config["submit_selector"])
        except Exception:
            # Now this is bad
            try:
                # Alternate fields that gmail might use
                await self.page.wait_for_selector(self.config["fall_back"]["password_selector"])
                await self.page.fill(self.config["fall_back"]["password_selector"], self.password)
                await self.page.click(self.config["submit_selector"])
            except Exception:
                return False

        return True
